/*******************************************************************************
**
** The header file for the Jahshaka core object file
** The Jahshaka Project
** Copyright (C) 2000-2004 The Jahshaka Project
** Released under the GNU General Public License
**
*******************************************************************************/

#ifndef COREOBJECT_H
#define COREOBJECT_H

#include "coreheaders.h"
#include "jahkeyframes.h"
#include "corestructures.h"
#include "videodata.h"          //for videodata
#include "projectdata.h"          //for videodata
#include <p_vector.h>                       //for particles


//for media support on layers
#include "mediaobject.h"
#include "assetexchange.h"
#include "jahpreferences.h"
#include "jahkeyframeobject.h"

////////////////////////////////////////////////////////////
//windows specific

#ifdef WIN32
#pragma warning (disable:4305) /* disable bogus conversion warnings */
#define drand48() (((float) rand())/((float) RAND_MAX))
#define lrand48() ((rand() << 16) ^ rand())
#define srand48(x) srand(x)
#endif

////////////////////////////////////////////////////////////
// the main layer object class

class JahLayer {

public:
    JahLayer(void);
    ~JahLayer(void);

    //////////////////////////////////////////////
    //declare generic functions first
    //////////////////////////////////////////////

    //initializes all the variables
    void setupDefaultVariables(void);
    
    //prints out all layer information
    void printInfo(void);

    // for gpu effects
    bool usedagpu;


    // core layer information
    ObjectCategory      objtype;
    LayerCategory       layertype;
    QString             layername;

    bool                layerStatus;    //means layer is on or off
    bool                selected;       //means layer is selected
    bool                drawtheLayer;   //used to hide layers even if on and selected
    bool                drawtheKey;     //used to toggle the keyasset
    bool                invertKey;      //used to invert the key in keyasset

    //////////////////////////////////////////////////////
    //resolution data for world layer
    /////////////////////////////////////////////////////

	void updateResolution(int res);

    //////////////////////////////////////////////////////
    //assets used to describe media connected to the layer
    /////////////////////////////////////////////////////

    //these could be made into pointers to be more efficient
    videoData           asset;
    videoData           keyasset;

    bool                layerAssetStatus;
    bool                keyAssetStatus;

    //assign a asset to a layer
    void assignAsset(videoData clip, VideoCategory cliptype);
    bool LoadLayerAsset( VideoCategory imagetype = CLIP );

    //assetexchange is replacing mediadata
    assetExchange*      mediaExchange;
    bool                mediaExchangePresent;

    void addMediaObj();

    //for video layers
    int                 currentframe;
    int                 inframe, outframe;

    //for keys
    int                 keycurrentframe;
    int                 keyinframe, keyoutframe;

    int                 keystartframe, keyendframe;

    //////////////////////////////////////////////
    //keyframe routines
    //////////////////////////////////////////////

    void setupKeyframeVariables(void);

    motionNode  *node;
    JahKeyframeObject * layernodes;
    
    void addNodes(int currentFrame, motionNode * newNode);
    void removeNodes(int currentFrame);
    void updateNodes(int currentFrame);  //gets interpolation
    void clear(void);
    void reset(void);

    //routine to draw keyframes for a layer
    void makeKeyFrames( void );
    bool drawKeyframes;

    ////////////////////////////////////////////////////
    // layer options
    bool option1;
    bool option2;
    bool option3;
    bool option4;

    ////////////////////////////////////////////////////
    //jahtexture texture calls here
    ////////////////////////////////////////////////////

    //layers can have keys
    TextureImage textureData;
    TextureImage textureKey;

    void makeTexture( void );
    void updateTexture( void );
    void makeLayerTexture( void );
    void makeKeyTexture( void );
    void changeKeyStatus(void );
    
    void rebuildTextureData   (  QImage & buf  );
    void rebuildTextureDataFx (  QImage & buf, bool conv  );
    void rebuildTextureKey    (  QImage & buf  );

    void updateVidTex( int theframe);
    bool updateVidTexAsset(int frame);
    bool updateVidTexKeyasset(int frame);

    void updateLayerTexture( void );
    void setLayerTexture(void);
    void setKeyTexture(void);

    void updateKeyTexture( void );

    TextureImage buildGlTexture (  QImage & buf  );
    void updateGlTexture ( QImage & buf );

    void InitializeTexture(TextureImage & textured);

    //return some layer data... used for new effects module
    //dont now if this is repetitious...
    int getImageWidth()  { return textureData.objectImage.width(); }
    int getImageHeight() { return textureData.objectImage.height(); }

    ///////////////////////////////////////////////
    //jahdraw calls here
    //////////////////////////////////////////////

    void            positionObject();
    void            drawLayer(void);

    ///////////////////////////////////////////////
    //generic jahobject object creation here
    ///////////////////////////////////////////////

    void            makeCylinder( void );
    void            makeSphere( void );
    void            makeLayer( void );
    void            makeKeyLayer( void );

    void            makeBackground( void );
    void            makeBlackBackground( void );
    void            makeBackdropOutline( void );
    void            makeCube( void );

    void            makeGrid( void );
    void            makeBezier(void);
    void            makeTitleGrid( void );
    
    ///////////////////////////////////////////////
    //generic jahobject object creation here
    ///////////////////////////////////////////////
    
    //outlines the main backdrop 
    void            makeLayerOutline( void); 

    //draws the onscreen selection widgets, should be moved out...
    void            makeLayerRotate( void);
    void            makeLayerTranslate( void);
    void            makeLayerScale( void);

    //////////////////////////////////////////////////
    //used for the light
    ////////////////////////////////////////////////////////

    void            makeLight( void);
    void            drawLight( void);


    //////////////////////////////////////////////////
    //used for layer fog
    ////////////////////////////////////////////////////////
    void            makeFog(void);
    bool            layerfog;

    //////////////////////////////////////////////////
    //used for the mesh layer
    ////////////////////////////////////////////////////////

    void            initializeMeshLayer( void);
    void            setMeshCoord( int i, int j, GLfloat x, GLfloat y);
    void            setMeshCoord( int i, int j, GLfloat x, GLfloat y, GLfloat z);
	void            getMeshCoord( int i, int j, GLfloat& x, GLfloat& y, GLfloat& z);
    void            makeMeshLayer( void );

    // can we use pointers here to free up memory?
    GLfloat         MeshXCoord [74][50]; // should be 72,48 for 720 486
    GLfloat         MeshYCoord [74][50];
    GLfloat         MeshZCoord [74][50];
    bool            MeshFirsttime;

    ////////////////////////////////////////////////////////
    //for the text object
    ////////////////////////////////////////////////////////

    textObj         *text;

    void            addTextObj();
    void            drawText();
    void            draw3dTextOutline();
    void            makeText(void );
    bool            loadFontText(void);

	bool			qtfont; //for qt gl fonts
	bool			loadQtFontText();


    ////////////////////////////////////////////////////
    //jah 3d model loading and drawing routines
    //////////////////////////////////////////////////

    QString         objfilename;
    QString         objextension;

    //////////////////////////////////////////////////
    //for glmlib objects
    bool            loadOBJ(const QString filename);
    void            makeObj(void);

    GLuint          model_list;                /* display list for object */
    GLMmodel*       model;                    /* glm model data structure */
    GLfloat         scale;                    /* original scale factor */
    GLfloat         smoothing_angle;          /* smoothing angle */
    char*           model_file;                /* name of the obect file */
    GLuint          material_mode;
    GLboolean       facet_normal;
	bool	        modelExists;

    //////////////////////////////////////////////////
    //for maxobjects
    bool            loadMaxOBJ(const QString filename);
    void            makeMaxObj(void);

    L3DS*           scene;
	bool	        sceneExists;

    ///////////////////////////////////////////////////
    //particle creation here
    //////////////////////////////////////////////////
    
    //this routine draws the partciles
    void            makeParticle();

    void            addParticleObj();

    void            Atom(bool first_time);
    void            JetSpray(bool first_time);
    void            Waterfall(bool first_time);
    void            Balloons(bool first_time);
    void            Rocket(bool first_time);
    void            Shape(bool first_time);
    void            FireFlies(bool first_time) ;
    void            Shower(bool first_time);
    void            Swirl(bool first_time);
    void            Explosion(bool first_time);
    void            Fountain(bool first_time);

    void            DrawGroupQuadSplat(const pVector &, const pVector &,
                                       float size_scale, bool draw_tex,
                                       bool const_size, bool const_color);

    void            DrawGroupTriSplat(const pVector &view, const pVector &up,
                                      float size_scale, bool draw_tex,
                                      bool const_size, bool const_color);

    //texturing routines used in particle engine
    inline float    Gaussian2(float x, float y, float sigma);
    void            MakeGaussianSpotTexture();  //used for particles

    //particle variables
    ParticleCategory    ParticleStyle;
    ParticleDrawstyle   ParticleDraw;
    bool                ParticleFirsttime;
    bool                ParticleColors;

    int             particle_handle, action_handle;
    int             maxParticles;
    int             demoNum,         numSteps,      prim;
    int             listID, SpotTexID;
    float           BlurRate;
    int             SteerShape;
    
    //GLuint      texture;

    //////////////////////////////////////////////////////////
    // For Vertex Shader object effects
    //////////////////////////////////////////////////

    void            initVertexShader(int shader);
    unsigned char * loadshaderfile(const char *name);

    // The vertex program
    GLuint          program;
    bool            pixelshader;

    //////////////////////////////////////////////////
    // screen resolution/ratio based variables for rendering
    //////////////////////////////////////////////////

    // should be in a subroutine later to be called whenever this data changes
    GLfloat         widthOffset;
    GLfloat         heightOffset;

	//this is a hack for now to let the layer know we are rendering
	//until we move the widgets out into the world object
	bool            renderStatus;

    ////////////////////////////////////////////////////
    //for effects layer support
    //////////////////////////////////////////////////

    bool            layerEffect;

    //these should be in a dynamic array that can be used
    //to add multiple effects to a layer
    //also should be node based so we can keyframe them

    bool            negative,swaprgb,mirror;
    bool            blur;
    GLfloat         hblur;
    GLfloat         vblur;

    //////////////////////////////////////////////////
    // global variables
    //////////////////////////////////////////////////

    //generic display settings
    GLenum      blendmode1, blendmode2;
    GLfloat     Red, Green, Blue;

    //generic layer flags and settings
    bool        blend;
    bool        depth;
    bool        cull;
    bool        lighting;
    bool        reflect;
    bool        smooth;
    bool        foreground;
    
    
    bool        ObjectDisplayMode;  //draw object as wireframe or as textured?
    bool        showEdges;          //used to hide or show head/tail for clips with slip
	bool		loop;
	bool		ping;


private:
    QString JahBasePath;
    QString JahMediaPath;
    
    int X_RESOLUTION;
    int Y_RESOLUTION;

};

#endif











































